#ifndef lint
static char rcsid[] = "$XConsortium: xclock.c,v 1.28 89/12/11 15:51:32 kit Exp $";
#endif /* lint */

/*
 * xclock --  Hacked from Tony Della Fera's much hacked clock program.
 *
 * xchrono -- Yet another hack by Nick Bender.
 */

/*
 * Copyright 1989 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * M.I.T. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL M.I.T.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>

#include "MultiClock.h"
#include <X11/Xaw/Cardinals.h>
#include "clock.bit"
#include "clmask.bit"

extern void exit();
static void quit();


static struct {
  char *timezone;
  char *name;
} cityInfo[] =
{
#ifdef SYSV
#define BOSTON 0
  { "EST5EDT", "Boston" },
#define NEWYORK (BOSTON+1)
  { "EST5EDT", "New York" },
#define CHICAGO (NEWYORK+1)
  { "CST6CDT", "Chicago" },
#define DENVER (CHICAGO+1)
  { "MST7MDT", "Denver" },
#define LA (DENVER+1)
  { "PST8PDT", "LA" },
#define HAWAII (LA+1)
  { "HST10", "Hawaii" },
#define TOKYO (HAWAII+1)
  { "JST-9", "Tokyo" },
#define SYDNEY (TOKYO+1)
  { "EST-10EDT", "Sydney" },
#define LONDON (SYDNEY+1)
  { "WET0WETDST", "London" },
#define PARIS (LONDON+1)
  { "MET-1METDST", "Paris" },
#define FRANKFURT (PARIS+1)
  { "MEZ-1MESZ", "Frankfurt" },
#define RIO (FRANKFURT+1)
  { "BST3", "Rio de Janeiro" },
#else /* Not SYSV */
#define BOSTON 0
  { "EST5EDT", "Boston" },
#define NEWYORK (BOSTON+1)
  { "EST5EDT", "New York" },
#define CHICAGO (NEWYORK+1)
  { "CST6CDT", "Chicago" },
#define DENVER (CHICAGO+1)
  { "MST7MDT", "Denver" },
#define LA (DENVER+1)
  { "PST8PDT", "LA" },
#define HAWAII (LA+1)
  { "HST", "Hawaii" },
#define TOKYO (HAWAII+1)
  { "Japan", "Tokyo" },
#define SYDNEY (TOKYO+1)
/* This should be NSW (New South Wales)? istead of Queensland,
   but that value looks wrong... 				*/
  { "Australia/Queensland", "Sydney" },
#define LONDON (SYDNEY+1)
  { "GB-Eire", "London" },
#define PARIS (LONDON+1)
  { "MET", "Paris" },
#define FRANKFURT (PARIS+1)
  { "MET", "Frankfurt" },
#define RIO (FRANKFURT+1)
  { "Brazil/East", "Rio de Janeiro" },
#endif /* SYSV */
};

static struct _resource {
  int updateInterval;
  int clockWidth;
  int clockHeight;
  String localName;
  Boolean boston;
  Boolean newyork;
  Boolean chicago;
  Boolean denver;
  Boolean la;
  Boolean hawaii;
  Boolean tokyo;
  Boolean sydney;
  Boolean london;
  Boolean paris;
  Boolean frankfurt;
  Boolean rio;
} resource;

#define offset(field)	XtOffset(struct _resource *, field)

static XtResource application_resources[] = {
  {"updateInterval", "UpdateInterval", XtRInt, sizeof (int),
     offset(updateInterval), XtRString, "60" },
  {"clockWidth", "ClockWidth", XtRInt, sizeof (int),
     offset(clockWidth), XtRString, "150" },
  {"clockHeight", "ClockHeight", XtRInt, sizeof (int),
     offset(clockHeight), XtRString, "200" },
  {"localName", "LocalName", XtRString, sizeof (String),
     offset(localName), XtRString, "" },
  {"boston", "Boston", XtRBoolean, sizeof(Boolean),
     offset(boston), XtRString, "false" },
  {"newyork", "Newyork", XtRBoolean, sizeof(Boolean),
     offset(newyork), XtRString, "false" },
  {"chicago", "Chicago", XtRBoolean, sizeof(Boolean),
     offset(chicago), XtRString, "false" },
  {"denver", "Denver", XtRBoolean, sizeof(Boolean),
     offset(denver), XtRString, "false" },
  {"la", "La", XtRBoolean, sizeof(Boolean),
     offset(la), XtRString, "false" },
  {"hawaii", "Hawaii", XtRBoolean, sizeof(Boolean),
     offset(hawaii), XtRString, "false" },
  {"tokyo", "Tokyo", XtRBoolean, sizeof(Boolean),
     offset(tokyo), XtRString, "false" },
  {"sydney", "Sydney", XtRBoolean, sizeof(Boolean),
     offset(sydney), XtRString, "false" },
  {"london", "London", XtRBoolean, sizeof(Boolean),
     offset(london), XtRString, "false" },
  {"paris", "Paris", XtRBoolean, sizeof(Boolean),
     offset(paris), XtRString, "false" },
  {"frankfurt", "Frankfurt", XtRBoolean, sizeof(Boolean),
     offset(frankfurt), XtRString, "false" },
  {"rio", "Rio", XtRBoolean, sizeof(Boolean),
     offset(rio), XtRString, "false" },
};

/* Command line options table.  Only resources are entered here...there is a
   pass over the remaining options after XtParseCommand is let loose. */

static XrmOptionDescRec options[] = {
{"-chime",	"*clock.chime",		XrmoptionNoArg,		"TRUE"},
{"-hd",		"*clock.hands",		XrmoptionSepArg,	NULL},
{"-hands",	"*clock.hands",		XrmoptionSepArg,	NULL},
{"-hl",		"*clock.highlight",	XrmoptionSepArg,	NULL},
{"-highlight",	"*clock.highlight",	XrmoptionSepArg,	NULL},
{"-update",	"*clock.update",	XrmoptionSepArg,	NULL},
{"-padding",	"*clock.padding",	XrmoptionSepArg,	NULL},
{"-d",		"*clock.analog",	XrmoptionNoArg,		"FALSE"},
{"-digital",	"*clock.analog",	XrmoptionNoArg,		"FALSE"},
{"-analog",	"*clock.analog",	XrmoptionNoArg,		"TRUE"},
{"-update",	"*updateInterval",	XrmoptionSepArg,	NULL},
{"-width", 	"*clockWidth",		XrmoptionSepArg,	NULL},
{"-height", 	"*clockHeight",		XrmoptionSepArg,	NULL},
{"-local",	"*localName",		XrmoptionSepArg,	NULL},
{"-boston",	"*boston",		XrmoptionNoArg,		"on"},
{"-newyork",	"*newyork",		XrmoptionNoArg,		"on"},
{"-chicago",	"*chicago",		XrmoptionNoArg,		"on"},
{"-denver",	"*denver",		XrmoptionNoArg,		"on"},
{"-la",		"*la",			XrmoptionNoArg,		"on"},
{"-hawaii",	"*hawaii",		XrmoptionNoArg,		"on"},
{"-tokyo",	"*tokyo",		XrmoptionNoArg,		"on"},
{"-sydney",	"*sydney",		XrmoptionNoArg,		"on"},
{"-london",	"*london",		XrmoptionNoArg,		"on"},
{"-paris",	"*paris",		XrmoptionNoArg,		"on"},
{"-frankfurt",	"*frankfurt",		XrmoptionNoArg,		"on"},
{"-rio",	"*rio",			XrmoptionNoArg,		"on"},
};


static XtActionsRec xclock_actions[] = {
    { "quit",	quit },
};

static Atom wm_delete_window;

/*
 * Report the syntax for calling xclock.
 */
Syntax(call)
	char *call;
{
	(void) printf ("Usage: %s [-analog] [-bw <pixels>] [-digital]\n", call);
	(void) printf ("       [-fg <color>] [-bg <color>] [-hd <color>]\n");
	(void) printf ("       [-hl <color>] [-bd <color>]\n");
	(void) printf ("       [-fn <font_name>] [-help] [-padding <pixels>]\n");
	(void) printf ("       [-rv] [-update <seconds>] [-display displayname]\n");
	(void) printf ("       [-geometry geom]\n");
	(void) printf ("       [-width clockWidth] [-height clockHeight] [-local localName]\n");
	(void) printf ("       [-boston] [-newyork] [-chicago] [-denver] [-la]\n");
	(void) printf ("       [-hawaii] [-tokyo] [-sydney] [-london]\n");
	(void) printf ("       [-paris] [-frankfurt] [-rio]\n\n");
	exit(1);
}

Arg clockArgs[] =
{
#define LABEL_ARG	0
  { XtNlabel, (XtArgVal) 0 },
#define TIMEZONE_ARG	(LABEL_ARG+1)
  { XtNtimezone, (XtArgVal) 0 },
#define WIDTH_ARG	(TIMEZONE_ARG+1)
  { XtNwidth, (XtArgVal) 0 },
#define HEIGHT_ARG	(WIDTH_ARG+1)
  { XtNheight, (XtArgVal) 0 },
};

void main(argc, argv)
    int argc;
    char **argv;
{
    Widget toplevel;
    Widget clockBox;
    Arg arg;
    Pixmap icon_pixmap = None;
    XtAppContext app_con;

    toplevel = XtAppInitialize (&app_con, "XClock", options, XtNumber(options),
				&argc, argv, NULL, NULL, ZERO);
    if (argc != 1) Syntax(argv[0]);

    XtAppAddActions (app_con, xclock_actions, XtNumber(xclock_actions));

    /* added (XtPoiner) Dave Brown 15/3/95;  brownd@csc.liv.ac.uk  */
    XtGetApplicationResources (toplevel, (XtPointer)&resource, application_resources,
			       XtNumber(application_resources), NULL, 0);

    /*
     * This is a hack so that f.delete will do something useful in this
     * single-window application.
     */
    XtOverrideTranslations(toplevel, 
		    XtParseTranslationTable ("<Message>WM_PROTOCOLS: quit()"));

    XtSetArg(arg, XtNiconPixmap, &icon_pixmap);
    XtGetValues(toplevel, &arg, ONE);
    if (icon_pixmap == None) {
	arg.value = (XtArgVal)XCreateBitmapFromData(XtDisplay(toplevel),
				       XtScreen(toplevel)->root,
				       clock_bits, clock_width, clock_height);
	XtSetValues (toplevel, &arg, ONE);
    }
    XtSetArg(arg, XtNiconMask, &icon_pixmap);
    XtGetValues(toplevel, &arg, ONE);
    if (icon_pixmap == None) {
	arg.value = (XtArgVal)XCreateBitmapFromData(XtDisplay(toplevel),
				       XtScreen(toplevel)->root,
				       clock_mask_bits, clock_mask_width, 
				       clock_mask_height);
	XtSetValues (toplevel, &arg, ONE);
    }


    clockBox = XtCreateManagedWidget ("clockBox", boxWidgetClass, toplevel, NULL, ZERO);

    clockArgs[WIDTH_ARG].value = (XtArgVal) resource.clockWidth;
    clockArgs[HEIGHT_ARG].value = (XtArgVal) resource.clockHeight;

    if (strlen(resource.localName))
      {
	clockArgs[LABEL_ARG].value = (XtArgVal) resource.localName;
	clockArgs[TIMEZONE_ARG].value = (XtArgVal) getenv("TZ");
	XtCreateManagedWidget ("local", clockWidgetClass, clockBox,
			       clockArgs, XtNumber(clockArgs));
      }

    if (resource.hawaii)
      {
	clockArgs[LABEL_ARG].value = (XtArgVal) cityInfo[HAWAII].name;
	clockArgs[TIMEZONE_ARG].value = (XtArgVal) cityInfo[HAWAII].timezone;
	XtCreateManagedWidget ("hawaii", clockWidgetClass, clockBox,
			       clockArgs, XtNumber(clockArgs));
      }

    if (resource.la)
      {
	clockArgs[LABEL_ARG].value = (XtArgVal) cityInfo[LA].name;
	clockArgs[TIMEZONE_ARG].value = (XtArgVal) cityInfo[LA].timezone;
	XtCreateManagedWidget ("la", clockWidgetClass, clockBox,
			       clockArgs, XtNumber(clockArgs));
      }

    if (resource.denver)
      {
	clockArgs[LABEL_ARG].value = (XtArgVal) cityInfo[DENVER].name;
	clockArgs[TIMEZONE_ARG].value = (XtArgVal) cityInfo[DENVER].timezone;
	XtCreateManagedWidget ("denver", clockWidgetClass, clockBox,
			       clockArgs, XtNumber(clockArgs));
      }

    if (resource.chicago)
      {
	clockArgs[LABEL_ARG].value = (XtArgVal) cityInfo[CHICAGO].name;
	clockArgs[TIMEZONE_ARG].value = (XtArgVal) cityInfo[CHICAGO].timezone;
	XtCreateManagedWidget ("chicago", clockWidgetClass, clockBox,
			       clockArgs, XtNumber(clockArgs));
      }

    if (resource.newyork)
      {
	clockArgs[LABEL_ARG].value = (XtArgVal) cityInfo[NEWYORK].name;
	clockArgs[TIMEZONE_ARG].value = (XtArgVal) cityInfo[NEWYORK].timezone;
	XtCreateManagedWidget ("newyork", clockWidgetClass, clockBox,
			       clockArgs, XtNumber(clockArgs));
      }

    if (resource.boston)
      {
	clockArgs[LABEL_ARG].value = (XtArgVal) cityInfo[BOSTON].name;
	clockArgs[TIMEZONE_ARG].value = (XtArgVal) cityInfo[BOSTON].timezone;
	XtCreateManagedWidget ("boston", clockWidgetClass, clockBox,
			       clockArgs, XtNumber(clockArgs));
      }

    if (resource.rio)
      {
	clockArgs[LABEL_ARG].value = (XtArgVal) cityInfo[RIO].name;
	clockArgs[TIMEZONE_ARG].value = (XtArgVal) cityInfo[RIO].timezone;
	XtCreateManagedWidget ("rio", clockWidgetClass, clockBox,
			       clockArgs, XtNumber(clockArgs));
      }

    if (resource.london)
      {
	clockArgs[LABEL_ARG].value = (XtArgVal) cityInfo[LONDON].name;
	clockArgs[TIMEZONE_ARG].value = (XtArgVal) cityInfo[LONDON].timezone;
	XtCreateManagedWidget ("london", clockWidgetClass, clockBox,
			       clockArgs, XtNumber(clockArgs));
      }

    clockArgs[LABEL_ARG].value = (XtArgVal) "GMT";
    clockArgs[TIMEZONE_ARG].value = (XtArgVal) "GMT";
    XtCreateManagedWidget ("gmt", clockWidgetClass, clockBox,
			   clockArgs, XtNumber(clockArgs));

    if (resource.paris)
      {
	clockArgs[LABEL_ARG].value = (XtArgVal) cityInfo[PARIS].name;
	clockArgs[TIMEZONE_ARG].value = (XtArgVal) cityInfo[PARIS].timezone;
	XtCreateManagedWidget ("paris", clockWidgetClass, clockBox,
			       clockArgs, XtNumber(clockArgs));
      }

    if (resource.frankfurt)
      {
	clockArgs[LABEL_ARG].value = (XtArgVal) cityInfo[FRANKFURT].name;
	clockArgs[TIMEZONE_ARG].value = (XtArgVal) cityInfo[FRANKFURT].timezone;
	XtCreateManagedWidget ("frankfurt", clockWidgetClass, clockBox,
			       clockArgs, XtNumber(clockArgs));
      }

    if (resource.tokyo)
      {
	clockArgs[LABEL_ARG].value = (XtArgVal) cityInfo[TOKYO].name;
	clockArgs[TIMEZONE_ARG].value = (XtArgVal) cityInfo[TOKYO].timezone;
	XtCreateManagedWidget ("tokyo", clockWidgetClass, clockBox,
			       clockArgs, XtNumber(clockArgs));
      }

    if (resource.sydney)
      {
	clockArgs[LABEL_ARG].value = (XtArgVal) cityInfo[SYDNEY].name;
	clockArgs[TIMEZONE_ARG].value = (XtArgVal) cityInfo[SYDNEY].timezone;
	XtCreateManagedWidget ("sydney", clockWidgetClass, clockBox,
			       clockArgs, XtNumber(clockArgs));
      }

    SetClockUpdateInterval (resource.updateInterval);
    XtRealizeWidget (toplevel);

    wm_delete_window = XInternAtom (XtDisplay(toplevel), "WM_DELETE_WINDOW",
				    False);
    (void) XSetWMProtocols (XtDisplay(toplevel), XtWindow(toplevel),
			    &wm_delete_window, 1);
    XtAppMainLoop (app_con);
}


static void quit (w, event, params, num_params)
    Widget w;
    XEvent *event;
    String *params;
    Cardinal *num_params;
{
    if (event->type == ClientMessage &&
	event->xclient.data.l[0] != wm_delete_window) {
	XBell (XtDisplay(w), 0);
	return;
    }
    XCloseDisplay (XtDisplay(w));
    exit (0);
}
